"""
Discobase - Relation database library using nothing but Discord.

Python Discord Codejam 2024
"""

from .__about__ import __license__, __version__
from .database import *
from .exceptions import *
from .table import *
